'' =================================================================================================
''   File....... Master Controller v8a.spin
''   Purpose.... Create routines for leg control
''   Author..... Bob Sweeney
''   Started.... 2/12/19
''--------------------------------------------------------------------------------------------------
''   1/15/19 Setup serial communications
''   1/17/19 Added routines for transmitting strings to selected legs (help from ersmith on Prop 1 forum)
''   1/18/19 Added IK routines and testing
''   1/27/19 Added IK movement for up and down
''   1/28/19 Added routine for narrow and wide stance
''   2/5/19  Added feedback comms from legs, v8a
'' =================================================================================================

con

  _xinfreq = 5_000_000
  _clkmode = xtal1 + pll16x

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  
con

  #1, HOME, #8, BKSP, TAB, CLREOL, CLRDN, #16      ' io formmatting control
  CR            = 13
  CLS           = 16
  Degrees       = (180.0/pi)
  Radians       = (pi / 180.0)

  
con
  RX1           = 31
  TX1           = 30
  SDA           = 29
  SCL           = 28

  'RGB LED pin setup values
  DAT_PIN       = 3
  CLK_PIN       = 4
  NUM_LED       = 24
  'RGB LED color IDs
  Red           = 0
  Green         = 1
  Blue          = 2
  Yellow        = 6
  White         = 4
  Orange        = 5
  LtBlue        = 3
  Off           = 7

'[BODY DIMENSIONS]  All units are in 1/10"
  CoxaLength      = 2.5        'Length of the Coxa [in]
  FemurLength     = 6.0        'Length of the Femur [in]
  TibiaLength     = 24.7       'Lenght of the Tibia [in]

obj
  io     : "pcFullDuplexSerial4FC" '1 COG for 4 serial ports
  RGB    : "LPD8806_20120731"
  F      : "F32_1_6.spin"
  FS     : "FloatString"
  numb   : "simple_numbers" 

var
  long comStack[100], outputstr
  long femurAngle, tibiaAngle, coxaAngle, Leg
  long IKfemurAngle, IKcoxaAngle, IKtibiaAngle, x1, y1, z1
  long LED_Data[NUM_LED], rgbstack[100]
  word fcolor, tcolor, ccolor, mcolor, blockoutput
  byte buffer[128], outbuf[128], colorbuf[32]

pub Start | n, angle, cognumber, x, z
  io.Init
  '  AddPort(vm_port,vm_rx, vm_tx, vm_cts,        vm_rts,        UARTS#DEFAULTTHRESHOLD, UARTS#NOMODE,UARTS#BAUD9600)
  'port 0 for comms to serial terminal
  io.AddPort(0,      31,    30,    io#PINNOTUSED, io#PINNOTUSED, io#DEFAULTTHRESHOLD, io#NOMODE, io#BAUD115200) 'Serial terminal output
  'port 1 for feedback comms from leg to master
  io.AddPort(1,      8,     9,     io#PINNOTUSED, io#PINNOTUSED, io#DEFAULTTHRESHOLD, io#NOMODE, io#BAUD115200)                                                      'IO port to master
  'port 2 - leg movement output via pin11
  io.AddPort(2,      10,     11,     io#PINNOTUSED, io#PINNOTUSED, io#DEFAULTTHRESHOLD, io#NOMODE, io#BAUD115200)                                                      'IO port to master
  io.Start

  cognew(GetColor, @rgbstack )

  F.start

  pauseMSec(1000)                                                                      'Start the ports

  io.tx(0,16)

  femurAngle := 320
  tibiaAngle := 0
  coxaAngle := 900

  x1 := 0.0        'coxa side to side value
  y1 := 23.0       'height - femur - 19.0 is minimum height, 30.0 max height, height is distance from leg tip to pivot point 
  z1 := 10.0        'tibia in and out value - (8.5 min to 9.0 max: 30.0 height)
                    '(4.0 min to 26.0 max at 19.0 height)- Distance is from pivot point to max leg extension outwards             
  n  := 3          'leg number

'  io.str(2,string("$,6,900,0,900",13))
'  SingleLegAngleMove(n)             'move selected leg using angle input
  SingleLegIK(n)                     'move single leg using position input - straight move to position
'  SingleLegIKincrementMove(n)       'move single leg from current position to new position using IK during movement
'  IKLegMoveTest                     'run values of x,z for compatable values, no actual leg movement
'  IKBodyDown                        'move body from full up position to full down position with legs in wide stance
'  IKBodyUp                          'move body from full down wide stance to full up position with legs in narrow stance
'  IKNarrowStance(n)                 'move all legs to a narrow stance with n being the front, maintain existing height
'  IKWideStance                      'move all legs to wide stance, maintain existing height
'  SetAllLegs                        'put all legs in position using position input
'  EvenLegMove
'  OddLegMove

pub SingleLegIK(legnum)
' move selected leg to x,y,z location using distances
  LegIK(x1,y1,z1)
  BuildString(legnum, IKFemurAngle, IKTibiaAngle, IKCoxaAngle)
  if blockoutput == false 
    io.str(2,@outbuf)
    io.str(0,@outbuf)
  else
    io.str(0,string("Bad output",TAB))
    io.str(0,@outbuf)

pub SingleLegAngleMove(legnum)
' move one leg to position using angles
  BuildString(legnum, FemurAngle, TibiaAngle, CoxaAngle)
  if blockoutput == false 
    io.str(2,@outbuf)
    io.str(0,@outbuf)
  else
    io.str(0,string("Bad output",TAB))
    io.str(0,@outbuf)

pub IKLegMoveTest | n,y,z
'used to check for bad IK ouput and mechanical interferences
    n := 3
    x1 := 0.0
    y1 := 5.0
    z1 := 5.0
    repeat z from 6 to 32
      y1 := 5.0
      repeat y from 5 to 32
        LegIK(x1,y1,z1)
        BuildString(n, IKFemurAngle, IKTibiaAngle, IKCoxaAngle)
'        if (ConstrainOutput == true) AND (InterferenceCheck == true)
          io.dec(0,z)
          io.tx(0,TAB)
          io.str(0,fs.FloatToFormat(y1,8,1))
          io.tx(0,TAB)
          io.str(0,fs.FloatToFormat(z1,8,1))
          io.tx(0,TAB)
          if ConstrainOutput == true
            io.str(0,string("Bad output",TAB))
          elseif InterferenceCheck == true
            io.str(0,string("Interference",TAB))
          io.str(0,@outbuf)
          y1 := f.fAdd(y1,1.0)
      z1 := f.fAdd(z1,1.0)

pub AllManualAngleLegMove
'move all legs to specified angles
    io.str(0,String(14))
    io.str(0,String("Start:"))
    io.str(2,string("$,1,900,0,900",13))
    io.str(2,string("$,2,900,0,900",13))
    io.str(2,string("$,3,900,0,900",13))
    io.str(2,string("$,4,900,0,900",13))
    io.str(2,string("$,5,900,0,900",13))
    io.str(2,string("$,6,900,0,900",13))
 
pub SetAllLegs | x
' set all legs to same position - direct movement, no load transfers, test stand only use
  repeat x from 1 to 6
    SingleLegIK(x)

pub ConstrainOutput
' validate output angles are within allowable values
  result := false
  if IKfemurAngle > FemurMaxAngle 
    result := true 
  elseif IKfemurAngle < FemurMinAngle
    result := true 
     
  if IKtibiaAngle > TibiaMaxAngle 
    result := true 
  elseif IKtibiaAngle < TibiaMinAngle
    result := true 
     
  if IKcoxaAngle > CoxaMaxAngle 
    result := true 
  elseif IKcoxaAngle < CoxaMinAngle
    result := true 
  
pub InterferenceCheck | x
' verify femur and tibia angles don't interfere with each other
  x := IKFemurAngle-IKTibiaAngle
  x := ||(x)
  if x =>250
    result := false
  else   
    result := true

PUB  LegIK (IKX, IKY, IKZ) | L1, L, C, IKb, IKb1, IKb2
'[LEG INVERSE KINEMATICS] Calculate motor angles for the given position of the feet
  L1 := f.fsqr(f.fadd(f.fmul(IKz,IKz),f.fmul(IKx,IKx)))
  L := f.fsub(L1, coxaLength)
  C := f.fsqr(f.fadd(f.fmul(IKy,IKy),f.fmul(L,L)))
  IKb := f.acos(f.fdiv((f.fsub((f.fadd((f.fmul(FemurLength,FemurLength)),(f.fmul(C,C)))),(f.fmul(TibiaLength,TibiaLength)))),(f.fmul((f.fmul(2.0,Femurlength)),C))))  
  IKb2 := f.atan2(L,IKy)
  IKb1 := f.acos(f.fdiv((f.fsub((f.fadd((f.fmul(tibiaLength,tibiaLength)),(f.fmul(C,C)))),(f.fmul(femurLength,femurLength)))),(f.fmul((f.fmul(2.0,tibialength)),C))))  

  IKfemurAngle := f.ftrunc(f.fmul(f.fmul(f.fadd(IKb,IKb2), Degrees),10.0))
  IKtibiaAngle := f.ftrunc(f.fmul(f.fmul(f.fsub(IKb2,IKb1), Degrees),10.0))
  IKcoxaAngle := f.ftrunc(f.fmul(f.fmul(f.atan2(IKz,IKx), Degrees),10.0))

  if (ConstrainOutput == true) or (InterferenceCheck == true)
    blockoutput := true


'--COLOR ROUTINES--------------------------------------------------------------------
  
pub ColorInitialize | n, i, j
  ' Start rgb driver and pass pointer to array.
  ' use variables fcolor, tcolor, ccolor to get color selections
  rgb.start(CLK_PIN, DAT_PIN, NUM_LED, @LED_Data)
  'test RGB output by cycling all lights
  repeat i from 0 to 5
    repeat j from 0 to 3
      n := (i * 4) + j
      LED_Data[n] := rgb.rgbColor(0, 127, 0) 'blue
      pauseMSec(25)
  'clear display
  repeat i from 0 to 5
    repeat j from 0 to 3
      n := (i * 4) + j
      LED_Data[n] := rgb.rgbColor(0, 0, 0)
      pauseMSec(25)

Pub SetLegColor(legnum, legitem, legcolor) | lednumber
' set specific LED color for a legitem
'leg: 1 - 6; leg 1 uses LED 0-3, leg 2 is LED 4-7, etc, leg 6 uses 20-23
'legitem: 1 - 4, 1- femur; 2- tibia; 3- coxa; 4- misc
'legcolor: 0 - 7; see constants for definitions
  lednumber := ((legnum-1) * 4) + (legitem - 1)            'figure out LED to lite
  case legcolor                                         'get color and push to LED
    0: LED_DATA[lednumber] := rgb.rgbColor(127, 0, 0) 'red
    1: LED_DATA[lednumber] := rgb.rgbColor(0, 127, 0) 'green
    2: LED_DATA[lednumber] := rgb.rgbColor(0, 0, 127) ' blue
    3: LED_DATA[lednumber] := rgb.rgbColor(127, 0, 127) ' LtBlue
    4: LED_DATA[lednumber] := rgb.rgbColor(127, 127, 127) 'white
    5: LED_DATA[lednumber] := rgb.rgbColor(127, 64, 0) 'orange
    6: LED_DATA[lednumber] := rgb.rgbColor(127, 127, 0) 'yellow
    7: LED_DATA[lednumber] := rgb.rgbColor(0, 0, 0) 'off

pub GetColor | n, color
  'set color for selected RGB LED
  ' cycle through all command items, decode and output RGB LED data
  ' check for feedback from legs
  ColorInitialize
  repeat
    repeat n from 1 to 6                                'cycle through legs
      bytefill(@buffer, 0, 64)                          'clear buffer
      colorbuf[0] := 0
      strColorAppend(@colorbuf,string("$,"))
      strColorAppend(@colorbuf,Numb.dec(n))
      strColorAppend(@colorbuf, string(13))
      io.str(1,@colorbuf)
      io.str(0,@colorbuf)
      
'      io.str(1,string("$,"))                            'output string: $,Legnumber,CR
'      io.str(1,Numb.dec(n))
'      io.str(1,string(13))           
      pauseMSec(25)                                     'allow time for response from leg - need better solution here...
      result := getcmd("$", @buffer, 62)                          'wait for response from leg
      if result > -1
        io.str(0,@buffer)
        io.str(0, String(13))
         
        SetLegColor(n, 1, extractfield(1, @buffer, ",")) 
        SetLegColor(n, 2, extractfield(2, @buffer, ",")) 
        SetLegColor(n, 3, extractfield(3, @buffer, ","))
        SetLegColor(n, 4, extractfield(4, @buffer, ","))
         

pub strColorAppend(buf, s)| n1, n2
  n1 := strsize(buf)
  n2 := strsize(s) + 1    'trailing 0
  buf += n1
  bytemove(buf, s, n2)  

pub buildstring(legnum, femur, tibia, coxa)
  outbuf[0] := 0
  strAppend(@outbuf, string("$,"))
  strAppend(@outbuf, Numb.dec(legnum))
  strAppend(@outbuf, string(","))
  strAppend(@outbuf, Numb.dec(femur))
  strAppend(@outbuf, string(","))
  strAppend(@outbuf, Numb.dec(tibia))
  strAppend(@outbuf, string(","))
  strAppend(@outbuf, Numb.dec(coxa))
  strAppend(@outbuf, string(13))

pub strAppend(buf, s)| n1, n2
  n1 := strsize(buf)
  n2 := strsize(s) + 1    'trailing 0
  buf += n1
  bytemove(buf, s, n2)  

pub getcmd(hdr, pstr, len) | idx, k, char                                             'routine designed by Jon McPhalen
' get command from serial stream
' hdr is required header character
' pstr is pointer to string space for command
' len is maximum length of input (buffer len-2)
  repeat
    char := io.rxcheck(1)
    if char == -1
      result := -1
      quit
    byte[pstr][0] := (k := char)                                             ' get char from stream
    if (k == hdr)                                                               ' if header
      quit
  if result == 0
    idx := 1                                                                      ' set starting index
    repeat while (idx < len)
      char := io.rxcheck(1)
      if char == -1
        quit
      byte[pstr][idx] := (k := char)                                           ' get char from stream
      if (k == 8)                                                                 ' if backspace
        idx := --idx #> 1                                                         ' backup
      elseif ((k == 0) OR (k == 13))                                              ' if 0 or CR, quit
        quit
      else
        ++idx                                                                     ' advance pointer
    byte[pstr][idx] := 0                                                          ' terminate string

pub extractfield(n, pstr, sep) | len, val, sign, flag, k                        'routine designed by Jon McPhalen
' returns decimal value of filed n
' pstr is source string/buffer
' sep is seperator character
   ' look for start of field
  len :=  strsize(pstr)                                                         ' maximum length to search
  repeat while (n > 0)
    if byte[pstr++] == sep                                                      ' if character is seperator
      --n                                                                       ' decrement field count
    if (--len == 0)                                                             ' if end of string
      return 0                                                                  ' early exit
  ' extract value    
  val := 0
  sign := 1
  flag := false
  repeat len
    k := byte[pstr++]                                                           ' get character from string
    case k
      " " : if (flag)                                                           ' if already in field
              quit
      "-" : ifnot (flag)                                                        ' if not in field 
              sign := -1                                                        ' set sign
            else
              quit  
      "0".."9" : val := (val * 10) + (k - "0")                                  ' convert ascii to decimal
                 flag := true                                                   ' mark entry to field                        
      other : quit                                                          
  return val * sign   

PUB pauseMSec(Duration)
'' Pause execution in milliseconds.
'' Duration = number of milliseconds to delay
  waitcnt(((clkfreq / 1_000 * Duration - 3932) #> 381) + cnt)


dat
femurMinAngle word 320
femurMaxAngle word 1500   

tibiaMinAngle word 0                                                            
tibiaMaxAngle word 520

coxaMinAngle word 290
coxaMaxAngle word 1510  


{{ Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}                  